/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __thread_Gang_h__
#define __thread_Gang_h__

namespace fe
{

/**************************************************************************//**
	@brief Group of Worker threads

	@ingroup thread
*//***************************************************************************/
template<typename WORKER,typename T>
class Gang: public JobQueue<T>
{
	public:
				Gang(void):
					m_ppWorker(NULL),
					m_ppThread(NULL),
					m_pGroup(NULL)

				{	Counted::setName("Gang"); }

virtual			~Gang(void)
				{
					if(JobQueue<T>::m_workers)
					{
						finish();
					}
				}

		BWORD	start(U32 a_threads,sp<Counted> a_spObject=sp<Counted>(NULL))
				{
					FEASSERT(!JobQueue<T>::m_workers);

					//* if threads arg is zero, create one thread per core
					if(!a_threads)
					{
						a_threads=Thread::hardwareConcurrency();
					}

					JobQueue<T>::m_workers=a_threads;
					sp< JobQueue<T> > spJobQueue(this);

					FEASSERT(!m_pGroup);
					m_pGroup=new Thread::Group();
					m_ppWorker=new WORKER*[a_threads];
					m_ppThread=new Thread*[a_threads];

					for(U32 m=0;m<a_threads;m++)
					{
						m_ppWorker[m]=new WORKER(m,spJobQueue);
						m_ppWorker[m]->setObject(a_spObject);
						m_ppThread[m]=NULL;
					}

					for(U32 m=0;m<a_threads;m++)
					{
						U32 tries=0;
						const U32 maxTries=3;
						while(TRUE)
						{
							//* TEMP
							static U32 serial=0;

							try
							{
								//* boost can throw boost::thread_resource_error
								m_ppThread[m]=m_pGroup->createThread(
										m_ppWorker[m]);

								serial++;
#if FE_CODEGEN<=FE_DEBUG
								if(!(serial%10000))
								{
									feLog("Gang::start thread serial %d\n",
											serial);
								}
#endif
							}
							catch (std::exception& e)
							{
								feLog("Gang::start"
										" failed to create boost thread "
										" worker %d/%d try %d/%d serial %d"
										" (\"%s\")\n",
										m,a_threads,tries,maxTries,
										serial,e.what());
								if(++tries>=maxTries)
								{
									feLog("Gang::start"
											" giving up creating thread\n");

									for(U32 n=0;n<m;n++)
									{
										feLog("Gang::start interrupt %d/%d\n",
												n,m);
										m_ppThread[n]->interrupt();
									}

									feLog("Gang::start cleaning up\n");
									discardWorkers();
									feLog("Gang::start done\n");

									return FALSE;
								}
								sleep(1);
								feLog("Gang::start retrying\n");
								continue;
							}
							if(tries)
							{
								feLog("Gang::start retry succeeded\n");
							}
							break;
						}
					}
					return TRUE;
				}

		void	finish(void)
				{
					FEASSERT(JobQueue<T>::m_workers);
					FEASSERT(m_pGroup);

					m_pGroup->joinAll();

					discardWorkers();

					//* empty out queue
					T job;
					while(take(job)){}
				}

	private:

		void	discardWorkers(void)
				{
					for(U32 m=0;m<JobQueue<T>::m_workers;m++)
					{
//						if(m_ppThread[m])
//						{
//							m_pGroup->remove_thread(m_ppThread[m]);
//							m_ppThread[m]=NULL;
//						}

						delete m_ppWorker[m];
						m_ppWorker[m]=NULL;
					}

					delete[] m_ppThread;
					m_ppThread=NULL;

					delete[] m_ppWorker;
					m_ppWorker=NULL;

					delete m_pGroup;
					m_pGroup=NULL;

					JobQueue<T>::m_workers=0;
				}

		Thread::Group*	m_pGroup;
		Thread**		m_ppThread;
		WORKER**		m_ppWorker;
};

} // namespace

#endif // __thread_Gang_h__
