/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __platform_dl_loader_h__
#define __platform_dl_loader_h__

namespace fe
{

/**************************************************************************//**
	@brief Raw access to a dynamic library

	@ingroup platform
*//***************************************************************************/
class FE_DL_PUBLIC DL_Loader
{
	public:
					DL_Loader(void)										{}

					/** @brief Open a dynamic library

						If adaptname is TRUE, the filename is adapted
						used using local convention
						(MyLib is interpreted as libMyLib.so or MyLib.lib). */
		BWORD		openDL(String filename,BWORD adaptname=TRUE);

					/// @brief Close a dynamic library
		BWORD		closeDL(void);

					/** @brief Find the named symbol in the open library

						Returns a pointer to the named function, if it exists.
						This may only work for functions wrapped in extern "C"
						and explicitly exported. */
		void		*findDLFunction(String symbol);

					/// @brief Returns the name of the library as it was opened
const	String&		name(void) const					{ return m_name; }

	private:

#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64
		HINSTANCE	m_handle;
#else
		void		*m_handle;
#endif

		String		m_name;
};

} // namespace

#endif // __platform_dl_loader_h__
